#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" `cli` focuses on the command line tool aspect using `click`

    .. seealso:: `Tasks`, `usage`, `clichain.pipeline`
"""

import click
import logging
import collections
from functools import wraps
import sys

from . import pipeline


logger = logging.getLogger(__name__)


class Tasks:
    """ Defines a factory to register tasks types.

        `Tasks` provides a factory to register and hold implemented
        tasks so they get available to the user.

        Tasks are implemented by coroutines.

        .. seealso:: `clichain.pipeline`

        The command line interface is implemented using `click`.
        In order to make a task usable from the command line interface,
        you need to define a `click` command (for most cases using
        `click.command` decorator).  example: ::

            @click.command(name='compute')
            @click.option('--approximate', '-a',
                          help='compute with approximation')
            @click.argument('x')
            def my_compute_task(approximate, x):
                " the task doc that will appear as help "
                # process inputs parameters and options...

        .. seealso:: `click` for details on how to implements commands

        The command is expected to return a coroutine function such as
        `clichain.pipeline.coroutine`, see `clichain.pipeline.create`
        for details.

        full example: ::

            from clichain import pipeline, cli
            import ast

            # creates the factory here but should be
            # common to all task types...
            tasks = cli.Tasks()

            @pipeline.task
            def divide(ctrl, den):
                print(f'{ctrl.name} is starting')

                with ctrl as push:
                    while True:
                        value = yield
                        push(value / den)

                print(f'{ctrl.name} has finished with no error')


            # the task will be made available as 'compute' in the
            # command line interface

            @tasks
            @click.command(name='compute')
            @click.option('--approximate', '-a',
                          help='compute with approximation')
            @click.argument('den')
            def compute_task_cli(approximate, den):
                " the task doc that will appear as help "
                if den == 0:
                    raise click.BadParameter("can't devide by 0")
                if approximate:
                    den = round(den)
                return devide(den)

            @pipeline.task
            def parse(ctrl):
                _parse = ast.literal_eval
                with ctrl as push:
                    while True:
                        try:
                            push(_parse((yield)))
                        except (SyntaxError, ValueError):
                            pass

            @tasks
            @click.command(name='parse')
            def parse_cli():
                " performs literal_eval on data "
                return parse()

        .. seealso:: `usage`
    """
    def __init__(self, commands=None):
        """ initializes new factory with `commands`

            `commands` is a dict containing all registered commands, and
            will be set to a new empty dict if None.

            `context` is the current click context, it's set to None
            until the main command is called (see `app`), which will
            set `context` to the current context value.
        """
        self.commands = {}
        self.context = None

    def __call__(self, cmd):
        """ wraps and register a `click.command` object into the factory

            `Tasks` object is intended to be used as a decorator: ::

                tasks = cli.Tasks()

                # register 'compute' command into 'tasks'
                @tasks
                @click.command(name='compute')
                [...]

            The command is expected to return a coroutine function such
            as `clichain.pipeline.coroutine`, see
            `clichain.pipeline.create` for details.

            .. note:: a log message will be emitted to indicate this
                task is created every time the command is called
        """
        cmdname = cmd.name
        self.commands[cmdname] = self._prepare_cmd(cmd)
        return cmd

    def _prepare_cmd(self, cmd):
        """ wraps the `click.command` callback function and replace it

            the wrapper function will:

            + log a 'create' message (using `logger.info`)

            + use the callback function result to create the next
              coroutine in the pipeline. The coroutine function created
              by the callback function is stored in the current `click`
              context. A stack is used to process pipeline's branches.

                .. seealso:: `clichain.pipeline` for details on how the
                    pipeline is specified

            The wrapper function does not return anything

            .. seealso:: this method is called by `Tasks.__call__`
        """
        f = cmd.callback

        @wraps(f)
        def callback(*args, **kw):
            logger.info(f'create: {cmd.name}')
            # NOTE: click.get_current_context() only works in the main
            #       thread, unless the current thread is run within
            #       the context called as a context manager.
            # XXX ctx = click.get_current_context().obj
            ctx = self.context.obj
            branch = ctx['branch']

            # create branch if first task in the branch
            if branch is None:
                branch = {'sequence': [], 'input': ctx['output']}
                ctx['branch'] = branch
                ctx['output'] = [ctx['index']]

            # add coroutine to current branch
            sequence = branch['sequence']
            coroutine = {
                'task': f(*args, **kw),
            }

            # set 'debug' name if required for the current branch
            if ctx['debug_name'] is not None:
                coroutine['debug'] = ctx['debug_name']

            sequence.append(coroutine)
            return None  # TODO should we return anything ?

        cmd.callback = callback
        return cmd


class Cli(click.MultiCommand):
    """ Implements root command using `click.MultiCommand`

        .. seealso:: `click`

        `Cli` provides implementation for the root command by extending
        `click.MultiCommand` (Then the click command is created
        specifying `Cli` as "cls" parameter in the `click.command`
        decorator.

        .. seealso:: `app`, `Tasks`
    """
    def list_commands(self, ctx):
        tasks = ctx.obj['tasks']
        rv = list(tasks.commands.keys())
        rv.append(_begin_fork.name)
        rv.append(_end_fork.name)
        rv.append(_new_branch.name)
        rv.append(_begin_debug_name.name)
        rv.append(_end_debug_name.name)
        return rv

    def get_command(self, ctx, name):
        tasks = ctx.obj['tasks']
        if name == _begin_fork.name:
            return _begin_fork
        elif name == _end_fork.name:
            return _end_fork
        elif name == _new_branch.name:
            return _new_branch
        if name == _begin_debug_name.name:
            return _begin_debug_name
        if name == _end_debug_name.name:
            return _end_debug_name
        try:
            return tasks.commands[name]
        except KeyError:
            raise click.UsageError(f'No such command "{name}"')


# NOTE: "\b" to prevent click from rewrapping examples paragraphs
def usage():
    """ create a pipeline of tasks, read text data from the
        standard input and send results to the standard output: ::

            stdin(text) --> tasks... --> stdout(text)

        The overall principle is to run a data stream processor by
        chaining different kinds of tasks (available tasks depending
        on the implementation, see list below).

        you can create a single branch pipeline as a sequence of
        tasks, for instance: ::

            inp >>> A -> B -> C >>> out

        or you can create a more complex pipeline defining multiple
        branches, for instance: ::

                       +--> B --> C --+
            inp >>> A--|              +--> F >>> out
                       +--> D --> E --+

        tasks are implemented by *coroutines* functions as described by
        David Beazle (see http://www.dabeaz.com/coroutines/ for details).



        + Specifying pipeline workflow:

          basic syntax allows you to specify the worflow of the pipeline.

          A single sequence of tasks as the following: ::

              inp >>> A -> B -> C >>> out

          is specified as: ::

              A B C

          .. note:: plus parameters and options of the tasks themselves,
              i.e: ::

                  A -x -y arg1 B -z ...

          Creating branches requires 'workflow commands', for instance
          the following example: ::

                         +--> B --> C --+
              inp >>> A--|              +--> F >>> out
                         +--> D --> E --+

          would be specified as: ::

              A [ B C , D E ] F

          the same way we can define sub branches, for instance: ::

                                  +--> C1 --+
                         +--> B --|         +-----+
                         |        +--> C2 --+     |
              inp >>> A--|                        +--> F >>> out
                         +--> D --> E ------------+

          would be specified as: ::

              A [ B [ C1 , C2 ] , D E ] F



        + Execution order:

          When *parallel* branches are defined (as 'C1' and 'C2' in the
          previous example) they are processed in the same order as they
          are defined in the command line arguments, that means in this
          example: ::

              A [ B [ C1 , C2 ] , D E ] F

          If the input data is: ::

              1
              2
              [...]

          Then the workflow will be such as: ::

              # data will go through C1 then C2
              1 -> A -> B -> C1 -> F
              1 -> A -> B -> C2 -> F
              2 -> A -> B -> C1 -> F
              2 -> A -> B -> C2 -> F
              [...]

          And the order is reproducible



        + Attaching a name to branches or tasks:

          You can attach a name to coroutines when defining the pipeline,
          which will be used as a suffix to get the logger if an
          exception occurs in the coroutne, i.e: ::

              base_logger.getChild(<name>)

          .. note:: This is useful in particular if you're using the
              same task type in several branches. The name could be used
              as well in the coroutine, depending on its implementation
              (see `clichain.pipeline.create` for more details).

          example: ::

                         +--> B --> C --+
              inp >>> A--|              +--> B >>> out
                         +--> B --> D --+

          you could specify the name of the branches (i.e all the
          coroutines of those branches) with: ::

              A [ { 'b1' B C } , { 'b2' B D } ] { 'b3' B )

          .. note:: the name specification is valid for every coroutine
              whose definition starts within the parenthesis,
              for example: ::

                  A { 'b1' [ B C , B D ] } B

              is equivalent to: ::

                  A [ { 'b1' B C , B D } ] B

              which is also equivalent to: ::

                  A [ { 'b1' B C , B D ] } B

              which is equivalent to: ::

                  A [ { 'b1' B C } , { 'b1' B D } ] B

              *note the output 'B' coroutine will have no name*

              And using the following specification: ::

                  A [ { 'b1' B } C , { 'b2' B } D ] B

              will only give 'b1' and 'b2' names to the 'B' coroutines
              (and not to the 'C' and 'D' coroutines as in the previous
              example).

              Then note the following: ::

                  A { 'b1' [ B C , { 'b2' B } D ] } B

              is equivalent to: ::

                  A [ { 'b1' B C } , { 'b2' B } { 'b1' D } ] B

              Then note that the following: ::

                  A [ { 'b1' } B , { 'b2' } B ] B

              will have no effect at all.
    """
    return _usage

_usage = usage.__doc__


@click.command(cls=Cli, chain=True,
               # add '\b' before code examples to avoid rewrapping
               # NOTE: add here because not rst compatible...
               help=usage().replace(' ::\n', ' ::\n\n\b'))
@click.option('--logfile', '-l',
              type=click.Path(exists=False),
              help='use a logfile instead of stderr')
@click.option('-v', '--verbose',
              count=True,
              help='set the log level: None=WARNING, -v=INFO, -vv=DEBUG')
@click.pass_context
def _app(ctx, logfile, verbose):
    """ this is the main command function, called by `app`

        the `_app` function itself only set up logging
    """
    log_levels = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }
    loglevel = log_levels.get(verbose, logging.DEBUG)
    # TODO more flexible logging config
    logging.basicConfig(format='%(name)s: %(levelname)s: %(message)s',
                        level=loglevel, filename=logfile)

    tasks = ctx.obj['tasks']
    tasks.context = ctx


@_app.resultcallback()
@click.pass_obj
def process(obj, rv, logfile, verbose):
    """ callback of the main command, called by `click`

        `process` creates the pipeline (using `clichain.pipeline.create`
        ), then run it with inputs from stdin and sending outputs to
        stdout (getting stdin and stdout from `click.get_text_stream`).

        if an exception occures then log the exception and raise
        `click.Abort`.

        .. seealso:: `clichain.pipeline`
    """
    _end_branch(obj)

    logger.info('creating pipeline...')
    stdin = click.get_text_stream('stdin')
    stdout = click.get_text_stream('stdout')

    def write(item):
        stdout.write(str(item))
        stdout.write('\n')

    pl = pipeline.create(obj['pipeline'], output=write)

    logger.info('processing...')
    logger.info('----------------------------------------')
    try:
        pl.run(stdin)
    except Exception as e:
        logger.error(f'abort: {type(e).__name__}: {e}')
        raise click.Abort from e
    logger.info('----------------------------------------')
    logger.info('DONE.')


@click.command(name=',')
@click.pass_obj
def _new_branch(obj):
    """ new branch """
    stack = obj['stack']
    head = stack[0]
    head['output'].extend(obj['output'])
    obj['output'] = head['input']
    _end_branch(obj)


@click.command(name='[')
@click.pass_obj
def _begin_fork(obj):
    """ begin fork """
    stack = obj['stack']
    stack.appendleft({
        'input': obj['output'],
        'output': [],
    })
    _end_branch(obj)


@click.command(name=']')
@click.pass_obj
def _end_fork(obj):
    """ end fork """
    stack = obj['stack']
    head = stack[0]
    head['output'].extend(obj['output'])
    obj['output'] = head['output']
    stack.popleft()
    _end_branch(obj)


def _end_branch(obj):
    br = obj['branch']
    if br is not None:
        index = obj['index']
        pl = obj['pipeline']

        # link and add branch's coroutines to the pipeline
        inp = br['input']
        # first to last-1 task in the branch
        for i, task in enumerate(br['sequence'][:-1]):
            task['input'] = inp
            key = f'{index}_{i}'
            pl[key] = task
            inp = key
        # last task in the branch
        # ( use <index> as key because it is the branch's output)
        task = br['sequence'][-1]
        task['input'] = inp
        pl[index] = task

        # prepare next branch
        obj['index'] = index + 1
        obj['branch'] = None


@click.command(name='{')
@click.argument('name')
@click.pass_obj
def _begin_debug_name(obj, name):
    """ begin debug name group """
    stack = obj['debug_name_stack']
    if obj['debug_name'] is not None:
        stack.appendleft(obj['debug_name'])
    obj['debug_name'] = name


@click.command(name='}')
@click.pass_obj
def _end_debug_name(obj):
    """ end debug name group """
    stack = obj['debug_name_stack']
    try:
        obj['debug_name'] = stack.popleft()
    except IndexError:
        obj['debug_name'] = None


def _get_obj(tasks, args, kw):
    """ TODO: useful for testing with `click.testing.CliRunner`

        >>> runner = CliRunner()
        >>> obj = cli._get_obj(tasks, (), {})
        >>> result = runner.invoke(cli._app, args, obj=obj)
    """
    output = ()
    return {
        'tasks': tasks,
        'pipeline': {},
        'stack': collections.deque({
            'input': [],
            'output': output,
        }),
        'debug_name_stack': collections.deque(),
        'debug_name': None,
        'index': 0,
        'args': args,
        'kwargs': kw,
        'branch': None,
        'output': output,
    }


def app(tasks, *args, **kw):
    """ run `click` main command: this will start the CLI tool.

        `tasks` is the `Tasks` factory to use (containing user commands)

        extra args and kwargs are added to the `click` context (**obj**)
        as '**args**' and '**kwargs**', they're not used by the
        framework.

        `app` uses `Cli` which extends `click.MultiCommand` to create
        the main command as a multicommand interface. This main command
        holds all the user defined commands and is the main entry point
        of the created tool.

        The pipeline itself is created and run by the `process`
        function, which is called when the main command itself returns,
        i.e when the all the input args have been processed by `click`.

        .. seealso:: `Tasks`, `process`

        .. seealso:: the main command itself only set up logging,
            see also `usage`
    """
    _app(obj=_get_obj(tasks, args, kw))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" `pipeline` module provides tools to create a pipeline of tasks

    a pipeline can be composed of one or several branches, but
    everything runs in a single thread. The initial goal of this
    framework is to provide a simple and direct way of defining task
    types and reuse them in different pipeline configurations.

    *The motivation is not to parallelise tasks yet tasks could be
    parallelized in some configurations, depending on the exact use
    case and the context...*

    tasks are implemented by *coroutines* functions as described by
    David Beazle (see http://www.dabeaz.com/coroutines/ for details).

    This module is used by `clichain.cli` module.
"""

from functools import wraps
import logging
import sys
import collections


logger = logging.getLogger(__name__)


def coroutine(func):
    """ coroutine decorator, 'prime' the coroutine function

        this function is intended to be used as a decorator to create
        a basic *coroutine* function, for instance: ::

            @coroutine
            def cr(*args, **kw):
                print('starting...')
                try:
                    while True:
                        item = yield
                        print(f'processing: {item}')
                except GeneratorExit:
                    print('ending...')

        calling the decorated function will automatically get it to the
        first *yield* statement. ::

            >>> cr()
            starting...

        .. note:: the decorated function is wrapped using
            `functools.wraps`

        .. seealso:: http://www.dabeaz.com/coroutines/
    """
    @wraps(func)
    def start(*args, **kwargs):
        cr = func(*args, **kwargs)
        next(cr)
        return cr
    return start


class Control:
    """ Internal, 'control' obj received by `task` decorated functions

        `Control` is a context manager

        .. seealso:: `Control.__init__`
    """
    def __init__(self, context, name, targets):
        """ initialize new `Control` object

            + `context` is a `Context` object, `Control` object
              will use it to log exception if an exception occurs
              in the coroutine while used as a context manager. The
              `Context` object is also accessible through the `context`
              attribute.

            + `name` will be accessible as `name` attribute (rw)

              ex: ::

                logger = logging.getLogger(f'{__name__}.{ctrl.name}')

            + `targets` will be accessible through `targets` property
              (ro)

              .. seealso:: `Control.targets`

            `targets` is read only to ensure consistency with `push`
            function returned by `Control.__enter__`: `Control` object
            is expected to be used as a context manager: ::

                with ctrl as push:
                    while True:
                        data = yield
                        push(data)  # send data to next coroutines

            a `push` function is defined and returned when `Control` is
            used as a context manager, but can actually be created using
            `Control.push` property.

            the purpose is to force using an efficient solution avoiding
            attributes lookup (using *self*) for every call, which has
            an impact given this function is likely to be called a lot
            (usually for each processed item). This way we define the
            function and reference it once in the user function (as
            'push' in the example).

            .. seealso:: `task` decorator
        """
        self.context = context
        self.name = name
        self._targets = targets

    @property
    def push(self):
        """ return a 'push' function sending data to next coroutines

            .. seealso:: `Control.__init__`
        """
        def push(value, targets=self.targets):
            """ send `value` to the following coroutines in the pipeline

                given 'targets' is an iterable containing the next
                coroutines, this is equivalent to:

                    for t in targets:
                        t.send(value)
            """
            for t in targets:
                t.send(value)

        return push

    def __enter__(self):
        """ return push function

            .. seealso:: `Control.push`, `Control.__exit__`
        """
        return self.push

    def __exit__(self, tpe, value, tb):
        """ handle GeneratorExit exception and log unhandled exceptions

            `Control` object is created by `task` decorator, the
            decorated function gets the `Control` object as first arg,
            and is expected to use it as a context manager, which will
            handle GeneratorExit and log any unhandled exception.

            `context` attribute (`Context` object) will be used if the
            exception is not `None` or `GeneratorExit`, in order to:

            + determine if the exception traceback should be logged,
              if the exception has already been logged by another
              `Control` object (i.e in another coroutine), then only
              an error message will be logged, otherwise the full
              exception will be recorded.

            + get the base logger to use

            .. seealso:: `task`, `Control.__init__`, `Context`
        """
        if isinstance(value, GeneratorExit):
            return True

        ctx = self.context
        lgr = ctx.logger.getChild(self.name)

        if value in ctx.exceptions:
            lgr.error(f'failing ({tpe.__name__}: {value})')
        else:
            lgr.error('an exception occured:', exc_info=(tpe, value, tb))
            ctx.exceptions.append(value)

    @property
    def targets(self):
        """ next coroutines in the pipeline """
        return self._targets


def task(func):
    """ make "partial" `coroutines` expected to be used with `create`.

        `task` will create a "partial" function, which when
        called with args and keyword args will actually return a
        `coroutine` function designed to be used with `create` function.

        example:

        a basic coroutine adding offset to input data could be defined
        as follows using `task`: ::

            @task
            def offset(ctrl, offset):
                print('pre-processing')

                with ctrl as push:
                    while True:
                        value = yield
                        push(value + offset)

                # will be executed unless an exception occurs in
                # the 'while' loop
                print('post_processing')

        + `ctrl` will handle the `GeneratorExit` exception and log any
          unhandled exception.

        + the `push` method send data to the next coroutines in the
          pipeline.

        the resulting function is called with the original function's
        args and keyword args: ::

            off = offset(offset=1)

        *off* is a partial `coroutine` function expected to be used in a
        pipeline defintion with `create`.

        the coroutine will eventually be created calling this new
        function with specific arguments depending on the pipeline
        specification (see `create` for details), ex: ::

            # create the coroutine
            off = off(targets=[t1, t2...])

        .. note:: as for `coroutine`, all the functions (partial or
            final functions) are wrapped using `functools.wraps`


        **example:**

        ::

            @task
            def output(ctrl):
                with ctrl:
                    while True:
                        print((yield))

            @task
            def parse(ctrl):
                with ctrl as push:
                    while True:
                        try:
                            value = ast.literal_eval((yield))
                        except (SyntaxError, ValueError):
                            continue
                        push(value)

            @task
            def offset(ctrl, offset):
                offset = int(offset)
                logger = logging.getLogger(f'{__name__}.{ctrl.name}')
                logger.info(f'offset: {offset}')

                with ctrl as push:
                    while True:
                        value = yield
                        push(value + offset)

                logger.info('offset task finished, no more value')

            if __name__ == '__main__':
                out = output()
                off1 = offset(10)
                off2 = offset(offset=100)
                parse = parse()

                # the previous results (out, off1, off2, proc) should
                # be used in the pipeline definition and the followings
                # should be performed by "create"
                out = out()
                off1 = off1((out,))
                off2 = off2((out,))
                parse = parse([off1, off2])

                with open('foo.txt') as inputs:
                    for data in inputs:
                        parse.send(data)

                out.close()
                off1.close()
                off2.close()
                parse.close()

        .. seealso:: `coroutine`, `create`
    """
    @coroutine
    @wraps(func)
    def _task(context, targets=None, debug='<noname>', args=None, kwargs=None):
        ctrl = Control(
            context=context,
            name=debug,
            targets=[] if targets is None else targets,
        )

        return func(ctrl,
                    *(() if args is None else args),
                    **({} if kwargs is None else kwargs))

    @wraps(func)
    def task(*args, **kwargs):
        @wraps(func)
        def _partial(*_args, **_kwargs):
            return _task(*_args, **_kwargs, args=args, kwargs=kwargs)
        return _partial

    return task


class Context:
    """ will be passed to all `coroutine`s in the pipeline

        `Context` object is a common object shared by all coroutines in
        the pipeline.

        attributes:

        + `exceptions` is a list which remains empty until an exception
          occurs within a `task` and is handled by the module. Then
          `exception` contains each exception caught by the module.
          Each exception is logged only one time with its traceback when
          it's caught by `Control` context manager.

          .. note:: if an exception is caught by the module it will be
            "re-raised" thus terminate the process but user code could
            still raise another exception(s) for example if a coroutine
            is not implemented using `task` or GeneratorExit is handled
            within the user loop...

        + `logger` will be used for every message logged by the module,
          and can be used by the user. The default is to use the
          module's `logger`.

        + `obj` attribute is an arbitrary object provided by the user
          when creating the pipeline. The default is `None`.

        .. seealso:: `create`
    """
    def __init__(self, logger=logger, obj=None):
        """ init the `Context` which will be shared by coroutines """
        self.logger = logger
        self.exceptions = []
        self.obj = obj


def create(tasks, output=print, **kw):
    """ create a pipeline of coroutines from a specification

        a pipeline is a succession of coroutines organized into one
        or several branches.

        `output` is a strategy to use for the pipeline output, the
        default strategy is `print`. `output` will be called for each
        data reaching the pipeline's output, it takes a single argument.

        **extra keyword args** will be used to initialize the `Context`
        object that will be send to all the coroutines of the pipeline.

        `tasks` argument describes the pipeline and consists of a
        mapping of coroutines as key: value pairs, where each single key
        identifies a single coroutine.

        each coroutine is defined either by a single `coroutine`
        function (see **task** field below) or a dictionnay, which
        contains the following fields:

        + **task**: the coroutine function to use

          the coroutine function will be called with the following
          keyword arguments:

          + **context**: the `Context` object shared by all the
            coroutines of the pipeline.

          + **targets**: an iterable containing the following coroutines
            in the pipeline (default is no targets).

          + **debug**: an optional name used to get a child logger from
            the `Context` logger, which will be used to log error if an
            exception occurs. The exception will be logged at error
            level and the exc_info will be passed to the log record.

            The value will be accessible (and writeable) through
            `Control.name` attribute, which can be usefull for logging:

            ex: ::

                logger = logging.getLogger(f'{__name__}.{ctrl.name}')

            .. note:: Default value is the coroutine's key in the
                pipeline definition (default will be used if value is
                `None` or an empty string).

        + **input**: (optional) set this coroutine as a 'target' of the
          coroutine(s) defined by **input**. **input** can be a single
          key or an iterable containing keys of other coroutines defined
          in the pipeline dictionnary.

          .. note:: `None` value will be interpreted as the pipeline's
            main input. No value or an empty list is equivalent as None
            if this coroutine is not specified as **output** of an other
            coroutine in the pipeline.

        + **output**: (optional) set the coroutine(s) whose keys are
          defined in **output** as a 'target' of this coroutine.
          **output** can be a single key or an iterable containing keys
          of other coroutines defined in the pipeline dictionnary.

          .. note:: `None` value will be interpreted as the pipeline's
            main output. No value or an empty list is equivalent as None
            if this coroutine is not specified as **input** of an other
            coroutine in the pipeline.

        + **debug**: (optional) a debug name to use in logging if an
          unhandled exception occurs. see above description.

        .. note:: specifying a coroutine by a `coroutine` function is
          equivalent as providing a dictionnary containing only the
          **task** field.


        **examples:**

        .. seealso:: `task`, `coroutine`

        given we have the following declarations: ::

            @coroutine
            def output(targets, **kw):
                try:
                    while True:
                        for t in targets:
                            t.send('RESULT: {}'.format((yield)))
                except GeneratorExit:
                    return

            @task
            def parse(ctrl):
                with ctrl as push:
                    while True:
                        try:
                            value = ast.literal_eval((yield))
                        except (SyntaxError, ValueError):
                            continue
                        push(value)

            @task
            def mytask(ctrl, param):
                logger = logging.getLogger(f'{__name__}.{ctrl.name}')
                logger.info('starting task')
                with ctrl as push:
                    while True:
                        [...]
                logger.info('finishing task')

        + defining a pipeline composed of a single sequence:

          example: ::

            inp >>> a --> b --> c --> d >>> out

          here's how we could define it: ::

            pipeline = pipeline.create({
                'a': parse(),
                'b': {'task': mytask(1), 'input': 'a'},
                'c': {'task': mytask(2), 'input': 'b'},
                'd': {'task': output, 'input': 'c'},
            })

          the created pipeline is a `Pipeline` object, it can be run
          over any input generator using its 'Pipeline.run' method,
          sending data to stdout by default.

          .. seealso:: `Pipeline.run`

        + define a pipeline with branches:

          example: ::

                       +--> B --> C >>> out
            inp >>> A--|
                       +--> D --> E >>> out

          here's how we could define it: ::

            pipeline = pipeline.create({
                'a': {'task': A, 'output': ('b', 'd')},
                'b': B,
                'd': D,
                'c': {'task': C, 'input': 'b'},
                'e': {'task': E, 'input': 'd'},
            })

          redoundant specification is not an issue, and the following
          example is equivalent to the previous one: ::

            pipeline = pipeline.create({
                'a': {'task': A, 'output': ('b', 'd')},
                'b': {'task': B, 'input': 'a', 'output': 'c'},
                'd': {'task': D, 'input': 'a', 'output': 'e'},
                'c': {'task': C, 'input': 'b', 'output': None},
                'e': {'task': E, 'input': 'd', 'output': ()},
            })

        + join branches

          example: given we want to implement this: ::

                       +--> B --> C --+
            inp >>> A--|              +--> N >>> out
                       +--> D --> E --+

          here's how we could define it: ::

            pipeline = pipeline.create({
                'a': {'task': A, 'output': ('b', 'd')},
                'b': B,
                'c': {'task': C, 'input': 'b', 'output': 'f'},
                'd': D,
                'e': {'task': E, 'input': 'd', 'output': 'f'},
                'f': F,
            })

        + control branches order

          the order in which coroutines are initialized, called and
          closed is reproducible.

          to control the data flow order between several branches just
          use the keys in the pipeline definition, as they will be
          sorted, like in the following example: ::

                       +--> (1) X --+
                       +--> (2) X --+
            inp >>> A--+--> (3) X --+--> B >>> out
                       +--> (4) X --+
                       +--> (5) X --+

          here's how we could define it: ::

            pipeline = pipeline.create({
                'a': A,
                1: {'task': X, 'input': 'a', 'output': 'b'},
                2: {'task': X, 'input': 'a', 'output': 'b'},
                3: {'task': X, 'input': 'a', 'output': 'b'},
                4: {'task': X, 'input': 'a', 'output': 'b'},
                5: {'task': X, 'input': 'a', 'output': 'b'},
                'b': B,
            })

          the 'X' coroutines will be initialized and processed in the
          expected order: 1, 2, 3, 4, 5 (they will be closed, if no
          exception occurs, in the opposite order).

        + loop back

          .. warning:: looping is currently not implemented and will
            raise a `NotImplementedError` when creating the pipeline.

          example: given we want to implement this: ::

                       +--> B --> C --+         + >>> out
            inp >>> A--|              +--> N -- +
                       +--> D --> E --+         |
                                      |         |
                             +--> F --+         |
                             |                  |
                             +------------------+

          here's how we could define it: ::

            pipeline = pipeline.create({
                'a': {'task': A, 'output': ('b', 'd')},
                'b': {'task': B, 'output': 'c'},
                'c': {'task': C},
                'd': {'task': D, 'output': 'e'},
                'e': {'task': E},
                'n': {'task': N, 'input': ('c', 'e'), 'output': None},
                'f': {'task': F, 'input': 'n', 'output': 'n'},
            })

          .. warning:: defining a loop can end up in an infinite recursion
              , no control is done on this, so it's up to the tasks
              implementation to handle this...

        + specify coroutines name

          in some contexts we may want to define a name for a coroutine
          which is different from its key.

          example: the previous example with ordered branches was: ::

                       +--> (1) X --+
                       +--> (2) X --+
            inp >>> A--+--> (3) X --+--> B >>> out
                       +--> (4) X --+
                       +--> (5) X --+

          here's how we could define it: ::

            pl = {
                'a': A,
                'b': B,
            }

            pl.update({
               i: {'task': X, 'input': 'a', 'output': 'b',
                   'debug': f"the X task number {i}"}
               for i in range(1, 6)
            })

            pl = pipeline.create(pl)
    """
    # ---------------------------------------------------------------- #
    # prepare coroutines
    # ---------------------------------------------------------------- #

    # NOTE: use OrderedDict to get reproducible outputs
    steps = collections.OrderedDict()  # <key>: <default>

    def default():
        return {
            'task': None,
            'input': set(),
            'output': set(),
        }

    for key, tsk in tasks.items():
        if key is None:
            raise ValueError('coroutine key cannot be None (reserved)')

        step = steps.setdefault(key, default())
        _inputs = _outputs = ()
        debug = str(key)

        if isinstance(tsk, collections.Mapping):
            _inputs = tsk.get('input', ())
            _outputs = tsk.get('output', ())
            debug = tsk.get('debug', debug)
            tsk = tsk['task']

        _inputs = _listify(_inputs)
        _outputs = _listify(_outputs)
        step['task'] = tsk
        step['debug'] = debug

        # by default if a coroutine has no input or no output the
        # pipeline input / output will be used

        # we also need to make sure coroutines are not added twice as
        # input / output of a given coroutine

        # in order to do that we link everything (except None)
        # using 'output':

        # iterate 'inputs' and add to 'input' only if None (i.e pipeline
        # input), else add to 'output' of the coroutine specified as
        # 'input'
        for inp in _inputs:
            if inp is None:
                step['input'].add(None)
            else:
                _in = steps.setdefault(inp, default())
                _in['output'].add(key)

        step['output'].update(_outputs)

        # debug checks
        assert key not in step['output'], 'coroutine output is itself'
        assert key not in step['input'], 'coroutine input is itself'

    # ---------------------------------------------------------------- #
    # create context
    # ---------------------------------------------------------------- #
    context = Context(**kw)

    # ---------------------------------------------------------------- #
    # prepare pipeline input / output
    # ---------------------------------------------------------------- #

    @coroutine
    def main_input(targets):
        try:
            while True:
                data = yield
                for target in targets:
                    target.send(data)
        except GeneratorExit:
            return

    @coroutine
    def main_output(strategy):
        try:
            while True:
                strategy((yield))
        except GeneratorExit:
            return

    main_output = main_output(output)
    pipeline = [main_output]

    # ---------------------------------------------------------------- #
    # link coroutines
    # ---------------------------------------------------------------- #

    lst = list(steps)
    loop = {}  # to detect loopback in the pipeline (not implemented)
    while lst:
        key = lst.pop(0)
        data = steps[key]
        output = data['output']

        # set default output if branch has no output
        targets = [main_output] if not output or None in output else []

        # NOTE: use sorted to get reproducible outputs
        for _out in sorted(output.difference({None}), key=str):
            # add the current coroutine as input of the target coroutine
            # so we know it has at least one input (otherwise the
            # pipeline input will be used, see below...)
            target = steps[_out]
            target['input'].add(key)

            if _out in lst:
                # the target coroutine has not been initialized yet
                # it will be initialized first
                break

            targets.append(target['task'])

        # skip for now if at least one target coroutine has not been
        # initialized yet
        if len(targets) < len(output):
            # detect loopback in the pipeline:
            # this is currently not implemented and will fail
            try:
                if loop[key] == len(lst):
                    raise NotImplementedError('recursion detected: looping is '
                                              'currently not implemented')
            except KeyError:
                pass
            loop[key] = len(lst)

            # skip and go back later to this one
            lst.append(key)
            continue

        # the coroutine can be initialized
        # ( NOTE: this will start the user generator function )
        tsk = data['task']
        cr = tsk(context=context, targets=targets, debug=data['debug'])
        data['task'] = cr
        pipeline.append(cr)

    # link pipeline input
    input_targets = []
    for key, data in steps.items():
        # link coroutine to pipeline input:
        # + if coroutine has no input, or
        # + if pipeline input (None) is specified in coroutine's inputs
        if not data['input'] or None in data['input']:
            input_targets.append(data['task'])

    pipeline.append(main_input(input_targets))
    pipeline.reverse()
    return Pipeline(pipeline)


def _listify(obj):
    """ makes sure `obj` is a `list` """
    if isinstance(obj, str):
        return [obj]
    try:
        # will fail if obj is not iterable
        return list(obj)
    except TypeError:
        return [obj]


class Pipeline:
    """ User interface returned by `create` function

        `Pipeline` object contains the `coroutine`s of a pipeline.

        When used as a context manager, it ensures that coroutines
        will be closed immediately at the end of the process.

        .. seealso:: `Pipeline.__enter__`, `Pipeline.__exit__`

        `Pipeline` also has an additional `Pipeline.run` method
        which can be called to run the pipeline over an input
        stream and wait until the process complete.
    """
    def __init__(self, targets):
        """ initialize a new pipeline with `coroutine`s

            `targets` is an iterable containing the coroutines of the
            pipeline, the first item must be the input coroutine.

            .. seealso:: `Pipeline.run`
        """
        self.targets = targets

    def run(self, inputs):
        """ run the pipeline over `inputs` iterator

            send data from `inputs` to the pipeline until their is no
            more data in inputs or an exception occurs.
        """
        with self as process:
            for data in inputs:
                process(data)

    def __enter__(self):
        """ return a function sending data thtough the pipeline

            ex: ::

                with pipeline as process:
                    for data in stream:
                        process(data)

            .. note:: this is equivalent to:

                ::

                    with pipeline:
                        target = pipeline.target
                        for data in stream:
                            target.send(data)

            .. seealso:: `Pipeline.__exit__`
        """
        def send(data, target=self.targets[0]):
            target.send(data)

        return send

    def __exit__(self, tpe, value, tb):
        """ close all the coroutines of the pipeline, raise exc if any

            The purpose of using the `Pipeline` object as a context
            manager is essentially to make sure all the coroutines will
            be terminated (closed) at the end of the process.

            This can be critical if user functions are expected to do
            some teardown jobs after processing data, for instance: ::

                # file won't be closed until the coroutine is closed
                # (see while loop...)

                @coroutine
                def cr(targets, *args, file, **kw):
                    with open(file) as f:
                        while True:
                            data = yield
                            [...]

            .. seealso:: `Pipeline.__enter__`
        """
        for coroutine in self.targets:
            coroutine.close()



#!/usr/bin/env python

import pytest
from clichain import cli, pipeline
import click
from click.testing import CliRunner
import sys
import ast
import logging


def test_main_help():
    tasks = cli.Tasks()
    runner = CliRunner()
    obj = cli._get_obj(tasks, (), {})
    args = ['--help']
    result = runner.invoke(cli._app, args, obj=obj)
    print('>> test_main_help #1:\n', result.output, file=sys.stderr)
    assert result.exit_code == 0
    assert not result.exception
    assert result.output.startswith(
"""Usage: _app [OPTIONS] COMMAND1 [ARGS]... [COMMAND2 [ARGS]...]...

  create a pipeline of tasks, """)

    # test examples formatting is correct (not rewrapped)
    assert """
                         +--> B --> C --+
              inp >>> A--|              +--> F >>> out
                         +--> D --> E --+
""" in result.output


def test_basic_single_task_help():
    tasks = cli.Tasks()
    
    @tasks
    @click.command(name='compute')
    @click.option('--approximate', '-a',
                  help='compute with approximation')
    @click.argument('y')
    def compute_task_cli(approximate, y):
        " divide data by 'y' "
        pass

    runner = CliRunner()
    obj = cli._get_obj(tasks, (), {})
    args = ['compute', '--help']
    result = runner.invoke(cli._app, args, obj=obj)
    print('>> test_basic_single_task_help #1:\n', result.output, file=sys.stderr)
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == """\
Usage: _app compute [OPTIONS] Y

  divide data by 'y'

Options:
  -a, --approximate TEXT  compute with approximation
  --help                  Show this message and exit.
"""


def test_basic_single_sequence():
    inputs = '1\n2\n3\n' 

    tasks = cli.Tasks()

    @pipeline.task
    def compute_task(ctrl, y):
        with ctrl as push:
            while True:
                push((yield) / y)

    @pipeline.task                                                      
    def parse(ctrl):                                                    
        _parse = ast.literal_eval                                       
        with ctrl as push:                                              
            while True:                                                 
                try:                                                    
                    push(_parse((yield)))                               
                except (SyntaxError, ValueError):                       
                    pass                                                
                                                                    
    @tasks                                                              
    @click.command(name='parse')                                        
    def parse_cli():                                                    
        " performs literal_eval on data "                               
        return parse() 

    @tasks
    @click.command(name='compute')
    @click.option('--approximate', '-a',
                  help='compute with approximation')
    @click.argument('y')
    def compute_task_cli(approximate, y):
        " divide data by 'y' "
        y = ast.literal_eval(y)
        if y == 0:
            raise click.BadParameter("can't devide by 0")
        if approximate:
            y = round(y)
        return compute_task(y)

    runner = CliRunner()
    obj = cli._get_obj(tasks, (), {})
    args = ['compute', '0']
    result = runner.invoke(cli._app, args, obj=obj,
                           input=inputs, catch_exceptions=False)
    print('>> test_basic_single_sequence #1:\n', result.output, file=sys.stderr)
    assert result.output == """\
Usage: _app compute [OPTIONS] Y

Error: Invalid value: can't devide by 0
"""

    obj = cli._get_obj(tasks, (), {})
    args = ['parse', 'compute', '10']
    result = runner.invoke(cli._app, args, obj=obj,
                           input=inputs, catch_exceptions=False)
    print('>> test_basic_single_sequence #2:\n', result.output, file=sys.stderr)
    assert result.output == """\
0.1
0.2
0.3
"""

def _get_pipeline(args, inputs, err_msg=None):
    tasks = cli.Tasks()
    
    @pipeline.task
    def compute_task(ctrl, name, err):
        print(f'starting {name} (coroutine name: {ctrl.name})')

        with ctrl as push:
            while True:
                value = yield
                if isinstance(value, str) and '7' in value and err:
                    raise RuntimeError(name if err_msg is None else err_msg)
                push(f'{value.strip()}->{name}')

        print(f'finished {name} (coroutine name: {ctrl.name})')
    
    @tasks
    @click.command(name='task')
    @click.argument('name')
    @click.option('--err', '-e', is_flag=True)
    def compute_task_cli(name, err):
        return compute_task(name, err)

    def test():
        runner = CliRunner()
        obj = cli._get_obj(tasks, (), {})
        result = runner.invoke(cli._app, args.split(), obj=obj,
                               input=inputs, catch_exceptions=False)
        return result

    return test


def test_pipeline_one_level_split_and_join():
    """
                   +--> B --> C --+                                         
        inp >>> A--|              +--> F >>> out                            
                   +--> D --> E --+  

    
        =>> A [ B C , D E ] F 
    """
    inputs = '\n'.join('12345678')
    args = 'task A [ task B task C , task D task E ] task F'
    name = 'test_pipeline_one_level_split_and_join'

    test = _get_pipeline(args, inputs)
    result = test()
    output = result.output
    print(f'>> {name} #1:\n', output, file=sys.stderr)
    assert output == """\
starting F (coroutine name: 3)
starting C (coroutine name: 1)
starting E (coroutine name: 2)
starting B (coroutine name: 1_0)
starting D (coroutine name: 2_0)
starting A (coroutine name: 0)
1->A->B->C->F
1->A->D->E->F
2->A->B->C->F
2->A->D->E->F
3->A->B->C->F
3->A->D->E->F
4->A->B->C->F
4->A->D->E->F
5->A->B->C->F
5->A->D->E->F
6->A->B->C->F
6->A->D->E->F
7->A->B->C->F
7->A->D->E->F
8->A->B->C->F
8->A->D->E->F
finished A (coroutine name: 0)
finished D (coroutine name: 2_0)
finished B (coroutine name: 1_0)
finished E (coroutine name: 2)
finished C (coroutine name: 1)
finished F (coroutine name: 3)
"""


def test_pipeline_two_levels_split_and_join():
    """
                            +--> C1 --+                                     
                   +--> B --|         +-----+                               
                   |        +--> C2 --+     |                               
        inp >>> A--|                        +--> F >>> out                  
                   +--> D --> E ------------+ 

        =>> A [ B [ C1 , C2 ] , D E ] F
    """
    inputs = '\n'.join('12345678')
    args = 'task A [ task B [ task C1 , task C2 ] , task D task E ] task F'
    name = 'test_pipeline_two_levels_split_and_join'

    test = _get_pipeline(args, inputs)
    result = test()
    output = result.output
    print(f'>> {name} #1:\n', output, file=sys.stderr)
    assert output == """\
starting F (coroutine name: 5)
starting C1 (coroutine name: 2)
starting C2 (coroutine name: 3)
starting E (coroutine name: 4)
starting B (coroutine name: 1)
starting D (coroutine name: 4_0)
starting A (coroutine name: 0)
1->A->B->C1->F
1->A->B->C2->F
1->A->D->E->F
2->A->B->C1->F
2->A->B->C2->F
2->A->D->E->F
3->A->B->C1->F
3->A->B->C2->F
3->A->D->E->F
4->A->B->C1->F
4->A->B->C2->F
4->A->D->E->F
5->A->B->C1->F
5->A->B->C2->F
5->A->D->E->F
6->A->B->C1->F
6->A->B->C2->F
6->A->D->E->F
7->A->B->C1->F
7->A->B->C2->F
7->A->D->E->F
8->A->B->C1->F
8->A->B->C2->F
8->A->D->E->F
finished A (coroutine name: 0)
finished D (coroutine name: 4_0)
finished B (coroutine name: 1)
finished E (coroutine name: 4)
finished C2 (coroutine name: 3)
finished C1 (coroutine name: 2)
finished F (coroutine name: 5)
"""


def test_pipeline_debug_name(caplog):
    """
                   +--> B --> C --+
        inp >>> A--|              +--> B >>> out
                   +--> B --> D --+

        #equivalent:
        =>> A { 'b1' [ B C , B D ] } B
        =>> A [ { 'b1' B C , B D } ] B
        =>> A [ { 'b1' B C , B D ] } B
        =>> A [ { 'b1' B C } , { 'b1' B D } ] B

        #
        =>> A [ { 'b1' B } C , { 'b1' B } D ] B

        #equivalent:
        =>> A { 'b1' [ B C , { 'b2' B } D ] } B
        =>> A [ { 'b1' B C } , { 'b2' B } { 'b1' D } ] B

        #no effect:
        =>> A [ { 'b1' } B , { 'b2' } B ] B
    """
    inputs = '\n'.join('12345678')
    name = 'test_pipeline_debug_name'

    def _test(ex, i, args, stdout, err_name='{NAME}'):
        print('', file=sys.stderr)
        err_name = err_name.format(NAME=f'{ex}_{i}')
        args = args.format(NAME=f'{ex}_{i}')
        test = _get_pipeline(args, inputs, err_msg=err_name)
        result = test()
        # abort:
        assert isinstance(result.exception, SystemExit)
        assert result.exit_code == 1

        print(f'\n>> {name}-{ex}#{i}: out:', file=sys.stderr)
        print(result.output, file=sys.stderr)
        assert result.output == stdout.format(NAME=f'{ex}_{i}')

        
        print(f'\n>> {name}-{ex}#{i}: logs:', file=sys.stderr)
        _logger = pipeline.logger.getChild(err_name)
        print(caplog.text, file=sys.stderr)
        assert any(map(lambda r: r.name == _logger.name and
                                 r.msg == "an exception occured:" and
                                 r.exc_info,
                       caplog.records))
        assert f"RuntimeError: {err_name}" in caplog.text

    # ---------------------------------------------------------------- #
    # equivalent:
    # =>> A { 'b1' [ B C , B D ] } B
    # =>> A [ { 'b1' B C , B D } ] B
    # =>> A [ { 'b1' B C , B D ] } B
    # =>> A [ { 'b1' B C } , { 'b1' B D } ] B
    # ---------------------------------------------------------------- #
    ex = 'ex_1'
    # ---------------------------------------------------------------- #
    args = [
        'task A {{ {NAME} [ task -e B task C , task B task D ] }} task B ',
        'task A [ {{ {NAME} task -e B task C , task B task D }} ] task B',
        'task A [ {{ {NAME} task -e B task C , task B task D ] }} task B',
        'task A [ {{ {NAME} task -e B task C }} , {{ {NAME} task B task D }} ] task B',
    ]

    stdout = """\
starting B (coroutine name: 3)
starting C (coroutine name: {NAME})
starting D (coroutine name: {NAME})
starting B (coroutine name: {NAME})
starting B (coroutine name: {NAME})
starting A (coroutine name: 0)
1->A->B->C->B
1->A->B->D->B
2->A->B->C->B
2->A->B->D->B
3->A->B->C->B
3->A->B->D->B
4->A->B->C->B
4->A->B->D->B
5->A->B->C->B
5->A->B->D->B
6->A->B->C->B
6->A->B->D->B
finished B (coroutine name: {NAME})
finished D (coroutine name: {NAME})
finished C (coroutine name: {NAME})
finished B (coroutine name: 3)
Aborted!
"""

    for i, args in enumerate(args):
        _test(ex, i, args, stdout)

    # ---------------------------------------------------------------- #
    # =>> A [ { 'b1' B } C , { 'b1' B } D ] B
    # ---------------------------------------------------------------- #
    ex = 'ex_2'
    # ---------------------------------------------------------------- #
    args = [
        'task A [ {{ {NAME} task -e B }} task C , {{ {NAME} task B }} task D ] task B',
    ]

    stdout = """\
starting B (coroutine name: 3)
starting C (coroutine name: 1)
starting D (coroutine name: 2)
starting B (coroutine name: {NAME})
starting B (coroutine name: {NAME})
starting A (coroutine name: 0)
1->A->B->C->B
1->A->B->D->B
2->A->B->C->B
2->A->B->D->B
3->A->B->C->B
3->A->B->D->B
4->A->B->C->B
4->A->B->D->B
5->A->B->C->B
5->A->B->D->B
6->A->B->C->B
6->A->B->D->B
finished B (coroutine name: {NAME})
finished D (coroutine name: 2)
finished C (coroutine name: 1)
finished B (coroutine name: 3)
Aborted!
"""

    for i, args in enumerate(args):
        _test(ex, i, args, stdout)

    # ---------------------------------------------------------------- #
    # equivalent:
    #  =>> A { 'b1' [ B C , { 'b2' B } D ] } B
    #  =>> A [ { 'b1' B C } , { 'b2' B } { 'b1' D } ] B
    # ---------------------------------------------------------------- #
    ex = 'ex_3'
    # ---------------------------------------------------------------- #
    args = [
        'task A {{ {NAME} [ task -e B task C , {{ {NAME}_bis task B }} task D ] }} task B ',
        'task A [ {{ {NAME} task -e B task C }} , {{ {NAME}_bis task B }} {{ {NAME} task D }} ] task B ',
    ]

    stdout = """\
starting B (coroutine name: 3)
starting C (coroutine name: {NAME})
starting D (coroutine name: {NAME})
starting B (coroutine name: {NAME})
starting B (coroutine name: {NAME}_bis)
starting A (coroutine name: 0)
1->A->B->C->B
1->A->B->D->B
2->A->B->C->B
2->A->B->D->B
3->A->B->C->B
3->A->B->D->B
4->A->B->C->B
4->A->B->D->B
5->A->B->C->B
5->A->B->D->B
6->A->B->C->B
6->A->B->D->B
finished B (coroutine name: {NAME}_bis)
finished D (coroutine name: {NAME})
finished C (coroutine name: {NAME})
finished B (coroutine name: 3)
Aborted!
"""

    for i, args in enumerate(args):
        _test(ex, i, args, stdout)

    # ---------------------------------------------------------------- #
    # =>> A [ { 'b1' } B , { 'b2' } B ] B
    # ---------------------------------------------------------------- #
    ex = 'ex_4'
    # ---------------------------------------------------------------- #
    args = [
        'task A [ {{ {NAME} }} task -e B , {{ {NAME} }} task B ] task B',
    ]

    stdout = """\
starting B (coroutine name: 3)
starting B (coroutine name: 1)
starting B (coroutine name: 2)
starting A (coroutine name: 0)
1->A->B->B
1->A->B->B
2->A->B->B
2->A->B->B
3->A->B->B
3->A->B->B
4->A->B->B
4->A->B->B
5->A->B->B
5->A->B->B
6->A->B->B
6->A->B->B
finished B (coroutine name: 2)
finished B (coroutine name: 3)
Aborted!
"""

    for i, args in enumerate(args):
        _test(ex, i, args, stdout, err_name='1')
